"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    }
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var Debugger = require("debug");
var qs = require("qs");
var isBuffer = require("is-buffer");
var methods_1 = require("./methods");
var model_1 = require("./model");
exports.Model = model_1.default;
var result_1 = require("./result");
var debug = Debugger('akita:client');
var INSTANCES = {};
function resolve(key) {
    if (!INSTANCES[key]) {
        INSTANCES[key] = create();
    }
    return INSTANCES[key];
}
function create(options) {
    var client = function client(path) {
        var _a;
        return _a = (function (_super) {
                __extends(AnonymousModel, _super);
                function AnonymousModel() {
                    return _super !== null && _super.apply(this, arguments) || this;
                }
                return AnonymousModel;
            }(model_1.default)),
            _a.path = path,
            _a.client = client,
            _a;
    };
    client._options = options || {};
    client._count = 0;
    client.create = create;
    client.resolve = resolve;
    client.setOptions = function (opts) {
        client._options = Object.assign({}, client._options, opts);
    };
    function getFormDataClass() {
        var FormData = client._options.FormData;
        if (!FormData && typeof window === 'object' && window.FormData) {
            FormData = window.FormData;
        }
        else if (!FormData && typeof global === 'object' && global.FormData) {
            FormData = global.FormData;
        }
        return FormData;
    }
    client.request = function request(path, init, query, reducer) {
        init = Object.assign({}, init);
        var queryParams = Object.assign({}, init.query);
        delete init.query;
        var queryString = qs.stringify(queryParams);
        if (queryString) {
            path += "?" + queryString;
        }
        if (init.headers) {
            init.headers = Object.assign({}, init.headers);
        }
        if (init.body && typeof init.body === 'object') {
            var body_1 = init.body;
            var FormData_1 = getFormDataClass();
            if (init.method === 'UPLOAD') {
                init.method = 'POST';
                if (!FormData_1) {
                    throw new Error('Akita Error: Can not resolve FormData class when use upload method');
                }
                if (!(body_1 instanceof FormData_1)) {
                    var form_1 = new FormData_1();
                    Object.keys(body_1).forEach(function (name) { return form_1.append(name, body_1[name]); });
                    init.body = form_1;
                }
            }
            else if (!isBuffer(body_1) && !(FormData_1 && body_1 instanceof FormData_1) && !(body_1 instanceof ArrayBuffer)) {
                if (!init.headers) {
                    init.headers = {};
                }
                if (!init.headers['Content-Type'] || init.headers['Content-Type'] === 'application/json') {
                    init.headers['Content-Type'] = 'application/json';
                    init.body = JSON.stringify(body_1);
                }
                else if (init.headers['Content-Type'] === 'application/x-www-form-urlencoded') {
                    init.body = qs.stringify(body_1);
                }
                else {
                    throw new Error("Akita Error: Unsupported Content-Type " + init.headers['Content-Type']);
                }
            }
        }
        if (client._options.init) {
            init = Object.assign({}, client._options.init, init);
            if (client._options.init.headers) {
                init.headers = Object.assign({}, client._options.init.headers, init.headers);
            }
        }
        if (!init.headers) {
            init.headers = {};
        }
        var apiRoot = client._options.apiRoot;
        if (apiRoot) {
            if (apiRoot[apiRoot.length - 1] === '/' && path[0] === '/') {
                path = path.substring(1);
            }
            else if (apiRoot[apiRoot.length - 1] !== '/' && path[0] !== '/') {
                path = "/" + path;
            }
            path = apiRoot + path;
        }
        if (debug.enabled) {
            debug(init.method, path, JSON.stringify(init));
        }
        var fetch = client._options.fetch;
        if (!fetch) {
            if (typeof window !== 'undefined') {
                fetch = window.fetch;
            }
            else if (typeof global !== 'undefined') {
                fetch = global.fetch;
            }
        }
        client._count += 1;
        return new result_1.default(fetch, path, init, query, reducer);
    };
    methods_1.default.forEach(function (method) {
        client[method] = function (path, init) {
            init = init || {};
            init.method = method.toUpperCase();
            return client.request(path, init);
        };
    });
    return client;
}
exports.default = resolve('default');
